(define-module (bytestructures guile explicit-endianness))
(import
 (bytestructures guile bytevectors)
 (bytestructures guile utils))
(include-from-path "bytestructures/body/explicit-endianness.scm")
(include-from-path "bytestructures/r7/explicit-endianness.exports.sld")
