(import
 (scheme base)
 (srfi 64)
 (bytestructures r7 utils)
 (bytestructures r7)
 (bytestructures r7 numeric-metadata)
 (bytestructures r7 bytevectors)
 (bytestructures r7 explicit-endianness))

(include "run-tests.body.scm")
